/*------------------------------------------------------------------------------
	Newton interface to the vCard parsing functions.
------------------------------------------------------------------------------*/

#include "Objects.h"
#include "ctype.h"


/*------------------------------------------------------------------------------
	F u n c t i o n   P r o t o t y p e s
------------------------------------------------------------------------------*/

extern		Ref	MakeBinary(RefArg rcvr, RefArg binClass, unsigned binSize);
extern		Ref	TextEncodingData(RefArg rcvr);
extern "C"	Ref	InstallCharToUnicodeEncoding(RefArg rcvr, RefArg charset);
extern		long	xConvertToUnicode(const void * inStr, UniChar * outStr, void * encoding, long numOfChars);
extern "C"	Ref	DecodeBase64(RefArg rcvr, RefArg inText);

extern "C"	Ref	DecodeCardItem(RefArg rcvr, RefArg ioContext);
static		char	AddParameter(RefVar & ioParams, char * inParamType, char * inParamValue);
inline		BOOL	IsWhiteSpace(char ch) { return ch == ' ' || ch == '\t'; }


/*------------------------------------------------------------------------------
	D e c o d i n g   F u n c t i o n s
------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------
	Parse a vCard/vCalendar line.
	The input is a chunk of unterminated ASCII text.
	NOTE	This data is destructively modified!

	The output is a frame of parsed items, of the form:
		{
			group:	'x,						any group identifier found
			params:	[['type,'dom],['type,'pref]],	parameters
			tag:		'EMAIL,					the command tag
			data:		"simon@localhost"		a string or binary object, decoded as necessary
		}

	Args:		ioContext
					frame containing data of the form:
					{
						data:		<asciString>,
						offset:	0
					}

	Return:	parsed frame
					nil => end of data
------------------------------------------------------------------------------*/

extern "C"
Ref DecodeCardItem(RefArg rcvr, RefArg ioContext)
{
	RefVar	group, tag, params, data;
	RefVar	result;

	if (!IsFrame(ioContext))
		ThrowBadTypeWithFrameData(kNSErrNotAFrame, ioContext);
	data = GetFrameSlot(ioContext, SYM(data));
	if (!IsBinary(data))
		ThrowBadTypeWithFrameData(kNSErrNotABinaryObject, data);

	RefVar	offsetSym(SYM(offset));
	long		offset = RefToInt(GetFrameSlot(ioContext, offsetSym));

	WITH_LOCKED_BINARY(data, dataPtr)
		char *	srcPtr = (char *) dataPtr + offset;
		char *	stopPtr = (char *) dataPtr + Length(data);
		char *	textPtr = srcPtr;
		char *	paramTypePtr = NULL;
		char *	dstPtr;
		char		ch;
		BOOL		isBase64 = NO;

		//	Decode command and parameters.
		while (srcPtr < stopPtr && (ch = *srcPtr++) != 0)
		{
			switch (ch)
			{
			case ' ':
			case '\t':
				if (textPtr == srcPtr - 1)
					textPtr = srcPtr;	// ignore whitespace before items
				else
					*(srcPtr-1) = 0;	// terminate the item early if whitespace after
				break;
			case '.':
				*(srcPtr-1) = 0;		// terminate this item
			//	item is group identifier
				group = MakeSymbol(textPtr);
				textPtr = srcPtr;		// reset for next item
				break;
			case '=':
				*(srcPtr-1) = 0;		// terminate this item
			//	item is parameter type [type, value, encoding...]
				paramTypePtr = textPtr;
				textPtr = srcPtr;		// reset for next item
				break;
			case ',':
				*(srcPtr-1) = 0;		// terminate this item
			//	item is parameter (in list)
				if (AddParameter(params, paramTypePtr, textPtr) == 'b')
					isBase64 = YES;
				textPtr = srcPtr;		// reset for next item
				break;
			case ';':
			case ':':
				*(srcPtr-1) = 0;		// terminate this item
				//	if no tag yet then this item is it
				//	else append this parameter to the list
				if (ISNIL(tag))
					tag = MakeSymbol(textPtr);
				else
				{
					if (AddParameter(params, paramTypePtr, textPtr) == 'b')
						isBase64 = YES;
					paramTypePtr = NULL;
				}
				textPtr = srcPtr;		// reset for next item
				if (ch == ':')
					srcPtr--;			// will force loop exit
				break;
			}
		}
		if (NOTNIL(tag))	// there was something in the line
		{
			result = AllocateFrame();
			SetFrameSlot(result, SYM(tag), tag);
			if (NOTNIL(group))
				SetFrameSlot(result, SYM(group), group);
			if (NOTNIL(params))
				SetFrameSlot(result, SYM(params), params);

			if (isBase64)
			{
				//	Decode remainder of (possibly very long folded) line as base64 data.
				// o	find end of data
				// o	create new binary object with clone of that data
				// o	data = DecodeBase64(thatData)
				BOOL	isNewLine = NO;
				unsigned	binSize;
				for (srcPtr = textPtr; srcPtr < stopPtr && (ch = *srcPtr) != 0; srcPtr++)
				{
					if (isNewLine && !IsWhiteSpace(ch))
						break;
					else
						isNewLine = (ch == 0x0A);
				}
				binSize = srcPtr - textPtr;
				RefVar	bin(MakeBinary(rcvr, SYM(binary), binSize));
				WITH_LOCKED_BINARY(bin, binPtr)
					BlockMove(textPtr, binPtr, binSize);
				END_WITH_LOCKED_BINARY(bin)
				SetFrameSlot(result, SYM(data), DecodeBase64(rcvr, bin));
			}
			else
			{
				//	Decode remainder of line as text.
				// First skip any leading space.
				while (textPtr < stopPtr && IsWhiteSpace(*textPtr))
					textPtr++;
				//	Unfold lines and substitute escaped characters.
				BOOL	isEscaped = NO;
				BOOL	isNewLine = NO;
				dstPtr = textPtr;
				for (srcPtr = textPtr; srcPtr < stopPtr && (ch = *srcPtr) != 0; srcPtr++)
				{
					if (isNewLine)
					{
						isNewLine = NO;
						if (!IsWhiteSpace(ch))
							break;
					}
					else if (isEscaped)
					{
						isEscaped = NO;
						switch (ch)
						{
						case 'n':
						case 'N':
							*dstPtr++ = 0x0D;
							break;
						case 't':
						case 'T':
							*dstPtr++ = 0x09;
							break;
						default:
							*dstPtr++ = ch;
						}
					}
					else if (ch == '\\')
						isEscaped = YES;
					else if (ch == 0x0A)
						isNewLine = YES;
					else if (ch != 0x0D)
						*dstPtr++ = ch;
				}
				// convert up to dstPtr to unicode
				unsigned actualSize;
				unsigned textSize = dstPtr - textPtr;
				RefVar str(MakeBinary(rcvr, SYM(string), (textSize + 1) * sizeof(UniChar)));
				RefVar encoding(TextEncodingData(rcvr));
				WITH_LOCKED_BINARY(encoding, encodingPtr)
					WITH_LOCKED_BINARY(str, strPtr)
						actualSize = xConvertToUnicode(textPtr, (UniChar *) strPtr, encodingPtr, textSize);
					END_WITH_LOCKED_BINARY(str)
				END_WITH_LOCKED_BINARY(encoding)

				if (actualSize != textSize)
					SetLength(str, (actualSize + 1) * sizeof(UniChar));
				SetFrameSlot(result, SYM(data), str);
			}
		}
		// Update the offset into the data.
		offset = srcPtr - (char *) dataPtr;
	END_WITH_LOCKED_BINARY(data)

	SetFrameSlot(ioContext, offsetSym, MAKEINT(offset));
	return result;
}


/*------------------------------------------------------------------------------
	Add a parameter to the parameter list.
	If the parameter is an encoding however, dont add it but return the
	encoding type.

	Args:		ioParams
					array object containing parameters decoded so far
				inParamType
					pointer to start of type text - could be NULL
				inParamValue
					pointer to start of value text

	Return:	letter indicating encoding
					'\0'	=>	no encoding specified
					'q'	=>	quoted-printable
					'b'	=>	base64
------------------------------------------------------------------------------*/

static char
AddParameter(RefVar & ioParams, char * inParamType, char * inParamValue)
{
	if (inParamType == NULL)
		inParamType = "type";
	else if (symcmp(inParamType, "encoding") == 0)
		return tolower(inParamValue[0]);

	RefVar	parm(AllocateArray(SYM(array), 2));
	SetArraySlot(parm, 0, MakeSymbol(inParamType));
	SetArraySlot(parm, 1, MakeSymbol(inParamValue));

	if (ISNIL(ioParams))
		ioParams = AllocateArray(SYM(array), 0);
	AddArraySlot(ioParams, parm);
	return 0;
}
