/*------------------------------------------------------------------------------
	Newton interface to the native text decoding functions.
------------------------------------------------------------------------------*/

#include "Objects.h"
#include "NewtonScript.h"
#include "Unicode.h"
#include "UStringUtils.h"
#include "ctype.h"
#include "UnicodeMapper.h"
#include "ObjHeader.h"


/*------------------------------------------------------------------------------
	F u n c t i o n   P r o t o t y p e s
------------------------------------------------------------------------------*/

extern		Ref	TextEncodingData(RefArg rcvr);
extern		long  AdjustLengthForEncoding(UniChar * inStr, long inLength, void * inData);
extern "C"	Ref	InstallCharToUnicodeEncoding(RefArg rcvr, RefArg charset);

extern		long	xConvertFromUnicode(const UniChar * inStr, void * outStr, void * encoding, long numOfChars);
extern		long	xConvertToUnicode(const void * inStr, UniChar * outStr, void * encoding, long numOfChars);

extern "C"	{
	Ref	EncodeHeader(RefArg rcvr, RefArg inText, RefArg inCoding);
	Ref	EncodeString(RefArg rcvr, RefArg inText, RefArg inCoding);
	Ref	Encode8bit(RefArg rcvr, RefArg inText);
	Ref	EncodeBase64(RefArg rcvr, RefArg inData, RefArg inStart, RefArg inCount, RefArg inForVCard);
	Ref	EncodeBase64String(RefArg rcvr, RefArg inText);

	Ref	DecodeHeader(RefArg rcvr, RefArg ioText);
	Ref	DecodeQuotedPrintableString(RefArg rcvr, RefArg inText);
	Ref	DecodeQuotedPrintable(RefArg rcvr, RefArg ioText);
	Ref	DecodeBase64String(RefArg rcvr, RefArg inText);
	Ref	DecodeBase64(RefArg rcvr, RefArg ioText);
	Ref	DecodeUUencode(RefArg rcvr, RefArg ioText);
	Ref	DecodeBinHex(RefArg rcvr, RefArg inText);

	Ref	UnicodeString(RefArg rcvr, RefArg inText);
	Ref	Reflow(RefArg rcvr, RefArg ioText);
}


/*------------------------------------------------------------------------------
	C o n s t a n t s
------------------------------------------------------------------------------*/

#define kRunArrayLength 64
#define kVBOThreshold (2*KByte)

#define kLeaveIt	MAKEINT(0)
#define kFlowIt	MAKEINT(1)
#define kQuoteIt	MAKEINT(2)


/*------------------------------------------------------------------------------
	U t i l i t i e s
------------------------------------------------------------------------------*/

Ref MakeBinary(RefArg rcvr, RefArg binClass, unsigned binSize)
{
	RefVar bin;
	if (binSize > kVBOThreshold)
	{
		RefVar store(NSSend(rcvr, SYM(GetDefaultOwnerStore)));
		bin = NSSend(store, SYM(NewVBO), binClass, MAKEINT(binSize));
	}
	else
	{
		bin = AllocateBinary(binClass, binSize);
	}
	return bin;
}


unsigned char ToHex(UniChar ch)
{
	unsigned x = (ch & 0x7F) - 0x30;	//	base on 0
	if (x > 9)
	{
		x -= 7;								//	handle A-F
		if (x > 0x20)						// handle lower-case
			x -= 0x20;
	}
	return x;
}


static inline UniChar ToUniChar(short x)
{
	UniChar ch = (x & 0x0F) + 0x30;	//	base on 0
	if (ch > '9')
		ch += 7;								//	make A-F
	return ch;
}


static inline unsigned StrLen(RefArg str)
{
	return Length(str) / sizeof(UniChar) - 1;
}


static long Ustrpos(const UniChar * str, const UniChar * str2, long start)
{
	const UniChar * s;

	for (s = str + start; *s != kEndOfUnicodeString; s++)
	{
		if (*s == *str2 && Ustrcmp(s, str2) == 0)
			return s - str;
	}
	return -1;
}


static int Ustrncmp(const UniChar * s1, const UniChar * s2, long n)
{
	int delta;
	while (*s1 && n-- > 0)
		if ((delta = (*s1++ - *s2++)) != 0)
			return delta;
	return 0;
}


/*------------------------------------------------------------------------------
	E n c o d i n g   F u n c t i o n s
------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------
	Encode RFC822 header string.

	Args:		inText
					the RFC822 header
				inCoding
					symbol of content transfer encoding

	Return:	a C string binary object - no NUL terminator
------------------------------------------------------------------------------*/

Ref EncodeHeader(RefArg rcvr, RefArg inText, RefArg inCoding)
{
	if (!IsString(inText))
		ThrowBadTypeWithFrameData(kNSErrNotAString, inText);
	if (!(ISNIL(inCoding) || IsSymbol(inCoding)))
		ThrowBadTypeWithFrameData(kNSErrNotASymbol, inCoding);

	// Build an array of runs to be encoded

	struct
	{
		unsigned short	start;
		unsigned short	count;
	} run[kRunArrayLength];
	unsigned numOfRuns = 0;

	WITH_LOCKED_BINARY(inText, textPtr)
		const UniChar * srcPtr = (UniChar *) textPtr;
		const UniChar * stopPtr = srcPtr + StrLen(inText);
		const UniChar * seqStartPtr = srcPtr, * seqStopPtr;
		bool isEncodable;
		UniChar ch;
		for ( ; srcPtr < stopPtr; srcPtr++)
		{
			ch = *srcPtr;
			if (ch == ' ' || ch == '"' || ch == '\'')
				seqStartPtr = srcPtr + 1;
			else if (ch >= 0x80)					//	non-ASCII character => word needs encoding
			{
				if (numOfRuns >= kRunArrayLength)
					ThrowBadTypeWithFrameData(kNSErrOutOfRange, MAKEINT(numOfRuns));
				run[numOfRuns].start = seqStartPtr - (UniChar *) textPtr;
				isEncodable = true;
				seqStopPtr = ++srcPtr;
				for ( ; srcPtr < stopPtr; srcPtr++)
				{
					ch = *srcPtr;
					if (ch == '\n' || ch == '"' || ch == '\'')
					{
						seqStopPtr = srcPtr;
						break;
					}
					else if (ch == ' ')
					{
						if (isEncodable)
						{
							isEncodable = false;
							seqStopPtr = srcPtr;
						}
						else
							break;
					}
					else if (ch >= 0x80)
						isEncodable = true;
				}
				run[numOfRuns].count = seqStopPtr - seqStartPtr;
				numOfRuns++;
			}
		}
	END_WITH_LOCKED_BINARY(inText)

	// Build a binary C string from the original header
	RefVar binStr(ASCIIString(inText));
	unsigned binEnd = Length(binStr);
	if (binEnd > 0)
		SetLength(binStr, binEnd - 1);	// dont want NUL terminator

	if (numOfRuns > 0)
	{
		bool isQP = EQ(inCoding, SYM(quoted-printable));
		// Build a preamble binary C string
		RefVar preamble;
		unsigned preambleLen;
		RefVar charset(GetVariable(rcvr, SYM(mimeCharset)));
		if (ISNIL(charset))
			ThrowBadTypeWithFrameData(kNSErrNotASymbol, charset);
		WITH_LOCKED_BINARY(charset, charsetPtr)
			char * srcPtr = (char *) charsetPtr + sizeof(ULong);	// SymbolName
			preambleLen = strlen(srcPtr) + 7;
			preamble = AllocateBinary(SYM(asciString), preambleLen);
			WITH_LOCKED_BINARY(preamble, preamblePtr)
				char * dstPtr = (char *) preamblePtr;
				*dstPtr++ = '=';
				*dstPtr++ = '?';
				while (*srcPtr)
					*dstPtr++ = *srcPtr++;
				*dstPtr++ = '?';
				*dstPtr++ = isQP ? 'Q' : 'B';
				*dstPtr++ = '?';
			// encoded string will go here
				*dstPtr++ = '?';
				*dstPtr   = '=';
			END_WITH_LOCKED_BINARY(preamble)
		END_WITH_LOCKED_BINARY(charset)

		// For each run,	copy text up to start to binary object
		//						copy encoding preamble
		//						copy encoded run
		//						copy postamble
		RefVar data;
		for (int i = numOfRuns - 1; i >= 0; i--)
		{
			data = Substring(inText, run[i].start, run[i].count);
			if (isQP)
			{
				data = EncodeString(rcvr, data, inCoding);
			}
			else
			{
				data = EncodeString(rcvr, data, NILREF);
				data = EncodeBase64(rcvr, data, MAKEINT(0), NILREF, NILREF);
			}
			BinaryMunger(binStr, run[i].start, run[i].count, preamble, 0, preambleLen);
			BinaryMunger(binStr, run[i].start + preambleLen - 2, 0, data, 0, -1);
		}
	}

	return binStr;
}


/*------------------------------------------------------------------------------
	Prepare a unicode string for mailing.
	Convert it to a C string; add LFs to CRs, provide dot transparency, flow it
	and optionally encode 8bit characters into quoted-printable representation.

	Args:		inText
					string

	Return:	a C string binary object - no NUL terminator
------------------------------------------------------------------------------*/

class CMailString
{
public:
				CMailString(const unsigned char * cStrPtr, unsigned cStrSize, RefArg inStr, unsigned char * inStrPtr, unsigned inStrSize);
	void		encode(bool isQP);
	unsigned	actualSize(void);

private:
	void		emit(unsigned char ch);

	RefStruct					str;
	unsigned char *			strPtr;
	unsigned						strSize;
	const unsigned char *	srcPtr;
	const unsigned char *	srcStopPtr;
	unsigned char *			dstPtr;
	unsigned char *			dstStopPtr;
	unsigned						lineLen;
};

const unsigned int kStretchIncr = 64;


CMailString::CMailString(const unsigned char * cStrPtr, unsigned cStrSize, RefArg inStr, unsigned char * inStrPtr, unsigned inStrSize)
{
	srcPtr = cStrPtr;
	srcStopPtr = cStrPtr + cStrSize;

	str = inStr;
	strPtr = inStrPtr;
	strSize = inStrSize;

	dstPtr = strPtr;
	dstStopPtr = strPtr + strSize - 1;
}


void CMailString::encode(bool isQP)
{
	unsigned char ch;
	const unsigned char * spacePtr = NULL;
	unsigned breakHere;
	bool isNewline = false;

	lineLen = 0;
	while (srcPtr < srcStopPtr)
	{
		ch = *srcPtr++;
		if (ch == '\n')
		{
			spacePtr = NULL;
			emit('\n');
			emit('\012');			// add LF to output
			lineLen = 0;			// reset line wrapping parameters
		}
		else if (ch == ' ' || ch == '\t')
		{
			emit(ch);
			spacePtr = srcPtr;	// remember where this space is
			breakHere = dstPtr - strPtr;
		}
		else if (ch == '.')
		{
			if (isNewline)
				emit('.');			// dot transparency for SMTP
			emit(ch);
		}
		else if (isQP && (ch == '=' || ch >= 0x80))
		{
			emit('=');				// encode quoted-printable character
			emit(ToUniChar(ch >> 4));
			emit(ToUniChar(ch));
		}
		else
		{
			emit(ch);
		}
		if (lineLen > 76)			// time to wrap line
		{
			if (spacePtr)
			{
				srcPtr = spacePtr;	// return pointers to last suitable break point
				dstPtr = strPtr + breakHere;
				spacePtr = NULL;
				if (!isQP)
				{
					emit('\n');				// flow it
					emit('\012');
					lineLen = 0;			// reset line wrapping parameters
				}
			}
			if (isQP)
			{
				emit('=');
				emit('\n');				// wrap it
				emit('\012');
				lineLen = 0;			// reset line wrapping parameters
			}
		}
		isNewline = (ch == '\n');
	}
}


void CMailString::emit(unsigned char ch)
{
	if (dstPtr >= dstStopPtr)
	{
		UnlockRefArg(str);
		SetLength(str, strSize + kStretchIncr);
		strPtr = (unsigned char *) LockedBinaryPtr(str);
		dstPtr = strPtr + strSize - 1;
		strSize += kStretchIncr;
		dstStopPtr = strPtr + strSize - 1;
	}
	lineLen++;
	*dstPtr++ = ch;
}


unsigned CMailString::actualSize()
{
	return dstPtr - strPtr;
}


/*------------------------------------------------------------------------------
	Make an 8-bit binary C string, possibly multi-byte encoded, from a unicode
	string, applying charset encoding.

	Args:		inText
					a string
				inCoding
					content transfer encoding symbol, or nil for no encoding

	Return:	a C string binary object - no NUL terminator
------------------------------------------------------------------------------*/

Ref EncodeString(RefArg rcvr, RefArg inText, RefArg inCoding)
{
	if (!IsString(inText))
		ThrowBadTypeWithFrameData(kNSErrNotAString, inText);

// apply charset encoding
	RefVar data(TextEncodingData(rcvr));
	unsigned actualSize;
	unsigned textSize = Length(inText) / sizeof(UniChar);
	RefVar cStr;

	WITH_LOCKED_BINARY(data, dataPtr)
		WITH_LOCKED_BINARY(inText, textPtr)
			cStr = MakeBinary(rcvr, SYM(asciString), AdjustLengthForEncoding((UniChar *) textPtr, textSize, dataPtr));
			WITH_LOCKED_BINARY(cStr, cStrPtr)
				actualSize = xConvertFromUnicode((const UniChar *) textPtr, cStrPtr, dataPtr, textSize);
			END_WITH_LOCKED_BINARY(cStr)
		END_WITH_LOCKED_BINARY(inText)
	END_WITH_LOCKED_BINARY(data)

	if (actualSize != textSize)
		SetLength(cStr, actualSize);	// dont add 1, dont want NUL terminator

// quoted-printable transfer encoding can only be done after charset encoding
	if (NOTNIL(inCoding))
	{
		SymbolObject * codingSymPtr = (SymbolObject *) RPTR(inCoding);
		bool  isQP = (codingSymPtr->hash == 0xC092F7A0);	// quoted-printable
		if (isQP || codingSymPtr->hash == 0xD03E2EE6			// 7bit
					|| codingSymPtr->hash == 0x6E75A89F)		// 8bit
		{
			// assume that on average lines will be broken at 64 chars and contain 2 quoted chars
			// => 7 extra chars / 64 or roughly 10% longer than the source
			// and create a binary object to receive this amount
			unsigned binStrSize = actualSize + actualSize/8;	// OK, 12% longer but faster math
			RefVar binStr(MakeBinary(rcvr, SYM(asciString), binStrSize));

			WITH_LOCKED_BINARY(cStr, cStrPtr)
				WITH_LOCKED_BINARY(binStr, binStrPtr)
					CMailString dst((const unsigned char *) cStrPtr, actualSize, binStr, (unsigned char *) binStrPtr, binStrSize);
					dst.encode(isQP);
					actualSize = dst.actualSize();
				END_WITH_LOCKED_BINARY(binStr)
			END_WITH_LOCKED_BINARY(cStr)

			SetLength(binStr, actualSize);
			cStr = binStr;
		}
	}

	return cStr;
}


/*------------------------------------------------------------------------------
	Make an 8-bit binary string from a unicode string.
	Ignore unicode mapping - just truncate to 8 bits.

	Args:		inText
					a string

	Return:	a C string binary object - no NUL terminator
------------------------------------------------------------------------------*/

Ref Encode8bit(RefArg rcvr, RefArg inText)
{
	if (!IsString(inText))
		ThrowBadTypeWithFrameData(kNSErrNotAString, inText);

	long cStrLen = Length(inText) / sizeof(UniChar);
	RefVar cStr(MakeBinary(rcvr, SYM(asciString), cStrLen));

	WITH_LOCKED_BINARY(inText, textPtr)
		WITH_LOCKED_BINARY(cStr, cStrPtr)
			xConvertFromUnicode((const UniChar *) textPtr, cStrPtr, NULL, cStrLen);
		END_WITH_LOCKED_BINARY(cStr)
	END_WITH_LOCKED_BINARY(inText)

	if (cStrLen > 0)
		SetLength(cStr, cStrLen - 1);	// dont want NUL terminator
	return cStr;
}


/*------------------------------------------------------------------------------
	Encode a binary object into base64 string representation.

	Args:		inData
					binary object or nil
				inStart
					index of first byte at which to start encoding
				inCount
					number of bytes to encode
					there are 19 groups of 3 = 57 bytes (76 characters) per line
					nil => encode to the end of the object
				inForVCard
					non-nil =>	fold each line by preceding it with a space
									and encode trailing = as q-p =3D

	Return:	a C string binary object - no NUL terminator
------------------------------------------------------------------------------*/

Ref EncodeBase64(RefArg rcvr, RefArg inData, RefArg inStart, RefArg inCount, RefArg inForVCard)
{
	if (ISNIL(inData))
		return NILREF;
	if (!IsBinary(inData))
		ThrowBadTypeWithFrameData(kNSErrNotABinaryObject, inData);
	if (!IsInt(inStart))
		ThrowBadTypeWithFrameData(kNSErrNotAnInteger, inStart);
	if (!(ISNIL(inCount) || IsInt(inCount)))
		ThrowBadTypeWithFrameData(kNSErrNotAnInteger, inCount);

	static const unsigned char tobase64[] =
	{
		'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z',
		'a','b','c','d','e','f','g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y','z',
		'0','1','2','3','4','5','6','7','8','9','+','/'
	};
	union
	{
		unsigned long asLong;
		unsigned char asChar[4];	// asChar[0] is never used
	} bits;
	unsigned start = RefToInt(inStart);
	unsigned count = ISNIL(inCount) ? Length(inData) - start : RefToInt(inCount);
	// the resulting string will be 4/3 x the binary size, padded up
	// plus 2 extra per line of 19 groups for CRLF
	unsigned grpCount = ((count + 2) / 3);
	unsigned binSize = grpCount * 4 + (grpCount / 19) * 2;
	if (NOTNIL(inForVCard))
	{
		long remainder = count % 3;
		if (remainder != 0)
			remainder = 3 - remainder;
		binSize += ((grpCount + 18) / 19		// for leading space on EVERY line
					+ remainder * 2);				// for encoding trailing = in quoted-printable
	}
	RefVar binStr(MakeBinary(rcvr, SYM(asciString), binSize));

	WITH_LOCKED_BINARY(inData, dataPtr)
		WITH_LOCKED_BINARY(binStr, binStrPtr)
			unsigned char * srcPtr = (unsigned char *) dataPtr + start;
			unsigned char * stopPtr = srcPtr + count - 3;
			unsigned char * dstPtr = (unsigned char *) binStrPtr;
			unsigned lineLen = 0;

			bits.asLong = 0;
			while (srcPtr <= stopPtr)
			{
				if (NOTNIL(inForVCard) && lineLen == 0)
					*dstPtr++ = ' ';

				bits.asChar[1] = *srcPtr++;		// encode 3 bytes...
				bits.asChar[2] = *srcPtr++;
				bits.asChar[3] = *srcPtr++;
				*dstPtr++ = tobase64[(bits.asLong >> 18) & 0x3F];	// ...into 4 chars
				*dstPtr++ = tobase64[(bits.asLong >> 12) & 0x3F];
				*dstPtr++ = tobase64[(bits.asLong >>  6) & 0x3F];
				*dstPtr++ = tobase64[ bits.asLong        & 0x3F];

				if (++lineLen >= 19)	// 19 groups of 4 = 76 characters per line
				{
					lineLen = 0;
					*dstPtr++ = 0x0D;
					*dstPtr++ = 0x0A;
				}
			}
			switch (srcPtr - stopPtr)
			{
			case 0:
			//	no padding required
				break;
			case 1:
				if (NOTNIL(inForVCard) && lineLen == 0)
					*dstPtr++ = ' ';
				bits.asLong = 0;
				bits.asChar[2] = *srcPtr++;		// encode 2 bytes and pad
				bits.asChar[3] = *srcPtr;
				*dstPtr++ = tobase64[(bits.asLong >> 10) & 0x3F];
				*dstPtr++ = tobase64[(bits.asLong >>  4) & 0x3F];
				*dstPtr++ = tobase64[(bits.asLong <<  2) & 0x3C];
				*dstPtr++ = '=';
				if (NOTNIL(inForVCard))
				{
					*dstPtr++ = '3';
					*dstPtr++ = 'D';
				}
				break;
			case 2:
				if (NOTNIL(inForVCard) && lineLen == 0)
					*dstPtr++ = ' ';
				bits.asLong = 0;
				bits.asChar[3] = *srcPtr;			// encode 1 byte and pad
				*dstPtr++ = tobase64[(bits.asLong >> 2) & 0x3F];
				*dstPtr++ = tobase64[(bits.asLong << 4) & 0x30];
				*dstPtr++ = '=';
				if (NOTNIL(inForVCard))
				{
					*dstPtr++ = '3';
					*dstPtr++ = 'D';
				}
				*dstPtr++ = '=';
				if (NOTNIL(inForVCard))
				{
					*dstPtr++ = '3';
					*dstPtr++ = 'D';
				}
				break;
			}
		END_WITH_LOCKED_BINARY(binStr)
	END_WITH_LOCKED_BINARY(inData)

	return binStr;
}


/*------------------------------------------------------------------------------
	Make a base64 encoded string from a unicode string.
	No charset encoding required - for authentication only.

	Args:		inText
					a string or nil

	Return:	a C string binary object - no NUL terminator
------------------------------------------------------------------------------*/

Ref EncodeBase64String(RefArg rcvr, RefArg inText)
{
	if (ISNIL(inText))
		return NILREF;
	if (!IsString(inText))
		ThrowBadTypeWithFrameData(kNSErrNotAString, inText);
	RefVar data(Encode8bit(rcvr, inText));
	data = EncodeBase64(rcvr, data, MAKEINT(0), NILREF, NILREF);

	unsigned dataSize = Length(data);
	RefVar str(MakeBinary(rcvr, SYM(string), (dataSize + 1) * sizeof(UniChar)));

	WITH_LOCKED_BINARY(data, dataPtr)
		WITH_LOCKED_BINARY(str, strPtr)
			xConvertToUnicode(dataPtr, (UniChar *) strPtr, NULL, dataSize);
		END_WITH_LOCKED_BINARY(str)
	END_WITH_LOCKED_BINARY(data)

	return str;
}


/*------------------------------------------------------------------------------
	D e c o d i n g   F u n c t i o n s
------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------
	Decode an envelope header.
	Decoding is done in place since the target string can never be longer than
	the source.
	ABNF:
	encoded-word = "=?" charset "?" encoding "?" encoded-text "?="
	charset = ISO-8859-1 | BIG5 
	encoding = B (base64) | Q (quoted-printable)
	eg encoded sequence is "=?ISO-8859-1?Q??=" | "=?BIG5?B??="

	Args:		ioText
					string containing quoted-printable escaped characters

	Return:	modified ioText
------------------------------------------------------------------------------*/

Ref DecodeHeader(RefArg rcvr, RefArg ioText)
{
	RefVar str;
	char  cstr[32], * pcstr;

	if (!IsString(ioText))
		return MakeString("");

	WITH_LOCKED_BINARY(ioText, textPtr)
		UniChar * srcPtr = (UniChar *) textPtr;
		UniChar * destPtr = srcPtr;

		while (*srcPtr)
		{
			if (*srcPtr++ == '=' && *srcPtr == '?')		// look for decodable candidate
			{
				UniChar * phraseStart = ++srcPtr - 2;
				UniChar * phraseEnd;
				UniChar   encodingStart;
				UniChar * codedStart;
				long      codedCount;

				pcstr = cstr;
				while (*srcPtr && *srcPtr != '?')			// copy charset
					*pcstr++ = *srcPtr++;
				*pcstr = 0;
				InstallCharToUnicodeEncoding(rcvr, MakeSymbol(cstr));

				encodingStart = tolower(*++srcPtr);
				while (*srcPtr && *srcPtr != '?')			// skip encoding
					srcPtr++;
				if (*srcPtr)
					srcPtr++;

				codedStart = srcPtr;
				while (*srcPtr)
				{
					if (*srcPtr++ == '?' && *srcPtr == '=')	// look for end of encoded phrase
					{
						codedCount = ++srcPtr - 2 - codedStart;
						str = Substring(ioText, codedStart - (UniChar *) textPtr, codedCount);
						if (encodingStart == 'q')
							str = DecodeQuotedPrintableString(rcvr, str);
						else if (encodingStart == 'b')
							str = DecodeBase64String(rcvr, str);
						else
							; // unknown encoding
						codedCount = StrLen(str);
						StrMunger(ioText, phraseStart - (UniChar *) textPtr, srcPtr - phraseStart, str, 0, codedCount);
						srcPtr = phraseStart + codedCount;
						break;
					}
				}
			}
		}
	END_WITH_LOCKED_BINARY(ioText)

	return ioText;
}


/*------------------------------------------------------------------------------
	Decode a quoted-printable unicode string.
	That means converting unicode to 8bit, decoding the q-p and converting
	back to unicode using the current charset.
	Also map underscores to spaces.
	NOTE	This function is destined solely for use by DecodeHeader.

	Args:		inText
					unicode string containing quoted-printable escaped characters

	Return:	new unicode string
------------------------------------------------------------------------------*/

Ref DecodeQuotedPrintableString(RefArg rcvr, RefArg inText)
{
// arg validation is done by Encode8bit

	RefVar str(Encode8bit(rcvr, inText));

	WITH_LOCKED_BINARY(str, strPtr)
		char * srcPtr = (char *) strPtr;
		for (int i = Length(str); i > 0; i--, srcPtr++)
		{
			if (*srcPtr == '_')
				*srcPtr = ' ';
		}
	END_WITH_LOCKED_BINARY(str)

	str = DecodeQuotedPrintable(rcvr, str);
	return UnicodeString(rcvr, str);
}


/*------------------------------------------------------------------------------
	Decode a quoted-printable binary string.
	Decoding is done in place since the target string can never be longer than
	the source.

	Args:		ioText
					binary unterminated string containing quoted-printable
					escaped characters

	Return:	modified ioText
------------------------------------------------------------------------------*/

Ref DecodeQuotedPrintable(RefArg rcvr, RefArg ioText)
{
	if (!IsBinary(ioText))
		ThrowBadTypeWithFrameData(kNSErrNotABinaryObject, ioText);

	unsigned originalSize = Length(ioText);
	unsigned actualSize;

	WITH_LOCKED_BINARY(ioText, textPtr)
		unsigned char * srcPtr = (unsigned char *) textPtr;
		unsigned char * stopPtr = srcPtr + originalSize;
		unsigned char * destPtr;
		unsigned char quotedChar;
		enum
		{
			kNothing,
			kQuote,
			kQuoted1
		} quotedState = kNothing;

		for ( ; srcPtr < stopPtr; srcPtr++)		// look for first candidate quote
		{
			if (*srcPtr == '=')
				break;
		}

		for (destPtr = srcPtr; srcPtr < stopPtr; srcPtr++)
		{
			switch (quotedState)
			{
			case kNothing:
				if (*srcPtr == '=')
					quotedState = kQuote;
			//	else if (*srcPtr == '_')
			//		*destPtr++ = ' ';
				else
					*destPtr++ = *srcPtr;
				break;

			case kQuote:
				if (*srcPtr == '\015')				// CR
				{
					if (*(srcPtr+1) == '\012')
						srcPtr++;						// skip LF
					quotedState = kNothing;			//	unfold "=\n"
				}
				else
				{
					quotedChar = ToHex(*srcPtr);
					quotedState = kQuoted1;
				}
				break;

			case kQuoted1:
				quotedChar = (quotedChar << 4) | ToHex(*srcPtr);
				*destPtr++ = quotedChar;
				quotedState = kNothing;
			}
		}
		*destPtr = kEndOfCharString;
		actualSize = destPtr - (unsigned char *) textPtr;
	END_WITH_LOCKED_BINARY(ioText)

	if (actualSize != originalSize)
		SetLength(ioText, (actualSize + 1) * sizeof(unsigned char));

	return ioText;
}


/*------------------------------------------------------------------------------
	Make a unicode string from a base64 encoded string.
	Arg validation is done by Encode8bit, DecodeBase64.

	Args:		inText
					string containing base64 encoded data

	Return:	a string
------------------------------------------------------------------------------*/

Ref DecodeBase64String(RefArg rcvr, RefArg inText)
{
	RefVar str(Encode8bit(rcvr, inText));
	str = DecodeBase64(rcvr, str);
	return UnicodeString(rcvr, str);
}


/*------------------------------------------------------------------------------
	Decode a binary string containing base64 characters into a binary object.
	Decoding is done in place since the target object can never be longer than
	the source.

	Args:		ioText
					binary unterminated string containing base64 characters

	Return:	a binary object - modified ioText
------------------------------------------------------------------------------*/

Ref DecodeBase64(RefArg rcvr, RefArg ioText)
{
	if (!IsBinary(ioText))
		ThrowBadTypeWithFrameData(kNSErrNotABinaryObject, ioText);

	static const signed char base64[] =
	{
		-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
		-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
		-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,62,-1,-1,-1,63,
		52,53,54,55,56,57,58,59,60,61,-1,-1,-1,00,-1,-1,
		-1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,
		15,16,17,18,19,20,21,22,23,24,25,-1,-1,-1,-1,-1,
		-1,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,
		41,42,43,44,45,46,47,48,49,50,51,-1,-1,-1,-1,-1
	};

	union
	{
		unsigned long asLong;
		unsigned char asChar[4];	// asChar[0] is never used
	} bits;
	unsigned actualSize = 0;

	WITH_LOCKED_BINARY(ioText, textPtr)
		unsigned char * srcPtr = (unsigned char *) textPtr;
		unsigned char * stopPtr = srcPtr + Length(ioText) - 4;
		unsigned char * dstPtr = srcPtr;
		long extra = 0;

		while (srcPtr <= stopPtr)
		{
			while (srcPtr < stopPtr && base64[*srcPtr] < 0)
				srcPtr++;										// look for valid base64 character
			if (srcPtr <= stopPtr)
			{
				bits.asLong = (base64[*srcPtr++] << 18)	// use next 4 chars...
								| (base64[*srcPtr++] << 12)
								| (base64[*srcPtr++] <<  6)
								|  base64[*srcPtr++];
				if (srcPtr > stopPtr)
				{
					if (*--srcPtr == '=')					// adjust length of final tuple
					{
						extra++;
						if (*--srcPtr == '=')
							extra++;
					}
				}
				*dstPtr++ = bits.asChar[1];				// ...to decode 3 bytes
				*dstPtr++ = bits.asChar[2];
				*dstPtr++ = bits.asChar[3];
			}
		}
		actualSize = (dstPtr - (unsigned char *) textPtr) - extra;
	END_WITH_LOCKED_BINARY(ioText)

	SetLength(ioText, actualSize);

	return ioText;
}


/*------------------------------------------------------------------------------
	Make a binary object from a uuencode encoded binary string.
	Decoding is done in place since the target object can never be longer than
	the source.

	Args:		ioText
					binary unterminated string containing uuencode encoded data

	Return:	a binary object - modified ioText
------------------------------------------------------------------------------*/

Ref DecodeUUencode(RefArg rcvr, RefArg ioText)
{
	if (!IsBinary(ioText))
		ThrowBadTypeWithFrameData(kNSErrNotABinaryObject, ioText);

	union
	{
		unsigned int  asLong;
		unsigned char asChar[3];
	} bits;
	unsigned actualSize;
	long start = -1;
	unsigned count = Length(ioText);

	//	Limit text within header and trailer
	//	begin 644 filename
	//	...
	//	end

	WITH_LOCKED_BINARY(ioText, textPtr)
		char * startPtr = (char *) textPtr;
		char * stopPtr;
		if (((startPtr = strstr(startPtr, "begin ")) != NULL)
		&& ((startPtr = strstr(startPtr + 6, "\015\012")) != NULL)
		&& ((stopPtr = strstr(startPtr + 2, "\015\012end")) != NULL)
		&& (stopPtr < startPtr + count))
		{
			start = startPtr - (char *) textPtr;
			count = stopPtr - startPtr;
		}
	END_WITH_LOCKED_BINARY(ioText)

	if (start < 0)
		return NILREF;

	WITH_LOCKED_BINARY(ioText, textPtr)
		unsigned char * srcPtr = (unsigned char *) textPtr + start;
		unsigned char * stopPtr = srcPtr + count - 4;
		unsigned char * dstPtr = srcPtr;
		unsigned char * lineStopPtr;

		while (srcPtr <= stopPtr)
		{
			lineStopPtr = srcPtr + ((*srcPtr++ - ' ') & 0x3F);
			while (srcPtr < lineStopPtr)
			{
			// handle partial block
				bits.asLong = ((*srcPtr++ - ' ') & 0x3F) << 18
								| ((*srcPtr++ - ' ') & 0x3F) << 12
								| ((*srcPtr++ - ' ') & 0x3F) <<  6
								| ((*srcPtr++ - ' ') & 0x3F);
				*dstPtr++ = bits.asChar[1];
				*dstPtr++ = bits.asChar[2];
				*dstPtr++ = bits.asChar[3];
			}
			while (srcPtr < stopPtr && *srcPtr < ' ')
				srcPtr++;
		}
		actualSize = dstPtr - ((unsigned char *) textPtr + start);
	END_WITH_LOCKED_BINARY(ioText)

	SetLength(ioText, actualSize);

	return ioText;
}


/*------------------------------------------------------------------------------
	Make a binary object from a binhex encoded binary string.
	Decoding is NOT done in place since the target object can be longer than
	the source which uses run-length compression.
	However, inText is modified.

	Args:		inText
					binary unterminated string containing uuencode encoded data

	Return:	a binary object
------------------------------------------------------------------------------*/

Ref DecodeBinHex(RefArg rcvr, RefArg inText)
{
	if (!IsBinary(inText))
		ThrowBadTypeWithFrameData(kNSErrNotABinaryObject, inText);

	static const signed char binhex[] =
	{
		-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
		-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
		-1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,-1,-1,
		13,14,15,16,17,18,19,-1,20,21,-1,-1,-1,-1,-1,-1,
		22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,-1,
		37,38,39,40,41,42,43,-1,44,45,46,47,-1,-1,-1,-1,
		48,49,50,51,52,53,54,-1,55,56,57,58,59,60,-1,-1,
		61,62,63,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1
	};

	union
	{
		unsigned int  asLong;
		unsigned char asChar[3];
	} bits;
	unsigned char filenameLength;
	unsigned filesize;
	unsigned originalSize = 0, actualSize = 0;
	long start = -1;
	unsigned count = Length(inText);

	//	Forget about the header:
	//	"(This file must be converted with BinHex 4.0)"
	//	and adjust start & count to the $: delimiters

	WITH_LOCKED_BINARY(inText, textPtr)
		char * startPtr = (char *) textPtr;
		char * stopPtr;
		if (((startPtr = strchr(startPtr, ':')) != NULL)
		&& ((stopPtr = strchr(startPtr + 1, ':')) != NULL)
		&& (stopPtr < startPtr + count))
		{
			start = startPtr - (char *) textPtr;
			count = stopPtr - startPtr;
		}
	END_WITH_LOCKED_BINARY(inText)

	if (start < 0)
		return NILREF;

	WITH_LOCKED_BINARY(inText, textPtr)
		unsigned char * srcPtr = (unsigned char *) textPtr + start;
		unsigned char * stopPtr = srcPtr + count - 4;
		unsigned char * dstPtr = srcPtr;
		long extra = 0;
		bool isRepCount = false;

	//	Pass 1 -- convert ASCII to binary

		while (srcPtr <= stopPtr)
		{
			while (srcPtr < stopPtr && binhex[*srcPtr] < 0)
				srcPtr++;										// look for valid binhex character
			if (srcPtr <= stopPtr)
			{
				bits.asLong = (binhex[*srcPtr++] << 18)	// use next 4 chars...
								| (binhex[*srcPtr++] << 12)
								| (binhex[*srcPtr++] <<  6)
								|  binhex[*srcPtr++];
			//	if (srcPtr > stopPtr)						// adjust length of final tuple?
				*dstPtr++ = bits.asChar[1];				// ...to decode 3 bytes
				*dstPtr++ = bits.asChar[2];
				*dstPtr++ = bits.asChar[3];
			}
		}
		originalSize = dstPtr - ((unsigned char *) textPtr + start) - extra;

	//	Pass 2 -- calculate decompressed size

		srcPtr = (unsigned char *) textPtr + start;
		stopPtr = srcPtr + originalSize;
		extra = 0;

		for ( ; srcPtr < stopPtr; srcPtr++)
		{
			if (isRepCount)
			{
				if (*srcPtr == 0)
					extra--;
				else
					extra += (*srcPtr - 3);
				isRepCount = false;
			}
			else
				isRepCount = (*srcPtr == 0x90);
		}
		actualSize = originalSize + extra;
	END_WITH_LOCKED_BINARY(inText)

	//	Pass 3 -- decompression

	RefVar bob(MakeBinary(rcvr, SYM(binary), actualSize));
	
	WITH_LOCKED_BINARY(inText, textPtr)
		WITH_LOCKED_BINARY(bob, bobPtr)
			unsigned char * srcPtr = (unsigned char *) textPtr + start;
			unsigned char * stopPtr = srcPtr + originalSize;
			unsigned char * dstPtr = (unsigned char *) bobPtr;
	
			for ( ; srcPtr < stopPtr; srcPtr++)
			{
				if (*srcPtr == 0x90)
				{
					unsigned char repCount = *srcPtr++;
					if (repCount == 0)
						*dstPtr++ = 0x90;
					else
					{
						unsigned char repValue = *(dstPtr - 1);
						while (--repCount > 0)
							*dstPtr++ = repValue;
					}
				}
				else
					*dstPtr++ = *srcPtr;
			}
			filenameLength = *(unsigned char *) bobPtr;
			filesize = *((unsigned char *) bobPtr + filenameLength + 12);
		END_WITH_LOCKED_BINARY(bob)
	END_WITH_LOCKED_BINARY(inText)

	/*
		And finally -- extract data fork from:
		Byte:    Length of FileName (1..63)
		Bytes:   FileName ("Length" bytes)
		Byte:    Version
		Long:    Type
		Long:    Creator
		Word:    Flags (AND $F800)
		Long:    Length of Data Fork
		Long:    Length of Resource Fork
		Word:    CRC
		Bytes:   Data Fork ("Data Length" bytes)
		Word:    CRC
		Bytes:   Resource Fork ("Rsrc Length" bytes)
		Word:    CRC
	*/
	BinaryMunger(bob, 0, filenameLength + 22, NILREF, 0, -1);
	BinaryMunger(bob, filesize, -1, NILREF, 0, -1);
	return bob;
}


/*------------------------------------------------------------------------------
	Convert a binary to a Unicode string, applying the current charset encoding.

	Args:		inText
					binary string to be converted

	Return:	a Unicode string
------------------------------------------------------------------------------*/

Ref UnicodeString(RefArg rcvr, RefArg inText)
{
	if (!IsBinary(inText))
		ThrowBadTypeWithFrameData(kNSErrNotABinaryObject, inText);

	RefVar data(TextEncodingData(rcvr));
	unsigned actualSize;
	unsigned textSize = Length(inText);
	RefVar str(MakeBinary(rcvr, SYM(string), (textSize + 1) * sizeof(UniChar)));

	WITH_LOCKED_BINARY(data, dataPtr)
		WITH_LOCKED_BINARY(inText, textPtr)
			WITH_LOCKED_BINARY(str, strPtr)
				actualSize = xConvertToUnicode(textPtr, (UniChar *) strPtr, dataPtr, textSize);
			END_WITH_LOCKED_BINARY(str)
		END_WITH_LOCKED_BINARY(inText)
	END_WITH_LOCKED_BINARY(data)

	if (actualSize != textSize)
		SetLength(str, (actualSize + 1) * sizeof(UniChar));

	return str;
}


/*------------------------------------------------------------------------------
	Reflow a unicode string - remove newlines after spaces.
	Do the transformation in place.

	Args:		ioText
					text to be reflowed

	Return:	modified ioText
------------------------------------------------------------------------------*/

Ref Reflow(RefArg rcvr, RefArg ioText)
{
	if (!IsString(ioText))
		ThrowBadTypeWithFrameData(kNSErrNotAString, ioText);

	unsigned actualSize = 0;

	WITH_LOCKED_BINARY(ioText, textPtr)
		UniChar * srcPtr = (UniChar *) textPtr;
		UniChar * destPtr = srcPtr;
		UniChar * sigPtr;
		bool spaced = false;
		UniChar ch;
		while (ch = *srcPtr++)	// intentional assignment
		{
			if (spaced && ch == '\n')	// ignore the newline
			{
	//			sigPtr = srcPtr - 5;
	//			if (sigPtr >= (UniChar *) textPtr && Ustrncmp(sigPtr, "\000\n\000-\000-\000 \000", 4) == 0)
	//				break;	// stop reflowing when sigdef reached
			}
			else
				*destPtr++ = ch;
			spaced = (ch == ' ');
		}
		*destPtr++ = kEndOfUnicodeString;

		if (srcPtr != destPtr)
			actualSize = (long) destPtr - (long) textPtr;
	END_WITH_LOCKED_BINARY(ioText)

	if (actualSize != 0)
		SetLength(ioText, actualSize);

	return ioText;
}
