/*------------------------------------------------------------------------------
	DIGEST-MD5 and CRAM-MD5 authentication algorithms.
------------------------------------------------------------------------------*/

#include "Objects.h"
#include "Unicode.h"
#include "UStringUtils.h"
#include "MD5.h"

#define kDigestLen     16
#define kMaxKeyLen     64
#define kMaxStringLen 255


/*------------------------------------------------------------------------------
	Convert binary to hex string.

	Args:		digest
					kDigestLen bytes
				cstring
					buffer for hex string

	Return:	void
------------------------------------------------------------------------------*/

void
ConvertToHexString(unsigned char * digest, unsigned char * cstring)
{
	const unsigned char hexDigits[] = "0123456789abcdef";
	unsigned char * cp = cstring;
	int i;
	for (i = 0; i < kDigestLen; i++)
	{
		unsigned char c = digest[i];
		*cp++ = hexDigits[c >> 4];
		*cp++ = hexDigits[c & 0x0F];
	}
	*cp = '\0';
}


/*------------------------------------------------------------------------------
	Newton interface to the MD5 Message Digest algorithm.

	Args:		string
					string ( < 256 characters) to be digested

	Return:	string of 32 hex characters
------------------------------------------------------------------------------*/

extern "C"
Ref Digest(RefArg rcvr, RefArg string)
{
	RefVar result;

	//	parameter must be a string
	if (IsString(string))
	{
		MD5_context   context;
		unsigned char digest[kDigestLen];
		unsigned char cstring[kMaxStringLen + 1];
		unsigned int  len;

		//	get length of string and convert Unicode -> ASCII
		WITH_LOCKED_BINARY(string, s)
			len = Ustrlen((UniChar *) s);
			ConvertFromUnicode((const UniChar *) s, (char *) cstring, kDefaultEncoding, kMaxStringLen);
			cstring[kMaxStringLen] = '\0';
		END_WITH_LOCKED_BINARY(string)

		//	digest
		MD5Init(&context);
		MD5Update(&context, cstring, len);
		MD5Final(digest, &context); 

		ConvertToHexString(digest, cstring);

		//	return Unicode string
		result = MakeString((const char *) cstring);
	}

	return result;
}


/*------------------------------------------------------------------------------
	Newton interface to the CRAM-MD5 authentication algorithm.

	Args:		string
					string ( < 256 characters) to be digested
				key
					string ( < 256 characters) of shared secret

	Return:	string of 32 hex characters
------------------------------------------------------------------------------*/

extern "C"
Ref CRAM(RefArg rcvr, RefArg string, RefArg key)
{
	RefVar result;

	//	parameters must be strings
	if (IsString(string) && IsString(key))
	{
		MD5_context context;
		unsigned char digest[kDigestLen];
		unsigned char cstring[kMaxStringLen + 1];
		unsigned char ckey[kMaxStringLen + 1];
		unsigned int  i, stringLen, keyLen;
		unsigned char k_ipad[kMaxKeyLen + 1];    // inner padding - key XORd with ipad
		unsigned char k_opad[kMaxKeyLen + 1];    // outer padding - key XORd with opad

		//	get length of string and convert Unicode -> ASCII
		WITH_LOCKED_BINARY(string, s)
			stringLen = Ustrlen((UniChar *) s);
			ConvertFromUnicode((const UniChar *) s, (char *) cstring, kDefaultEncoding, kMaxStringLen);
			cstring[kMaxStringLen] = '\0';
		END_WITH_LOCKED_BINARY(string)

		//	get length of key and convert Unicode -> ASCII
		WITH_LOCKED_BINARY(key, s)
			keyLen = Ustrlen((UniChar *) s);
			ConvertFromUnicode((const UniChar *) s, (char *) ckey, kDefaultEncoding, kMaxStringLen);
			ckey[kMaxStringLen] = '\0';
		END_WITH_LOCKED_BINARY(string)

		/* if key is longer than 64 bytes reset it to key = MD5(key) */
		if (keyLen > kMaxKeyLen)
		{
			MD5Init(&context);
			MD5Update(&context, ckey, keyLen);
			MD5Final(ckey, &context);
			keyLen = kDigestLen;
		}

		/*
		 * the HMAC_MD5 transform looks like:
		 *
		 * MD5(K XOR opad, MD5(K XOR ipad, text))
		 *
		 * where K is an n byte key
		 * ipad is the byte 0x36 repeated 64 times

		 * opad is the byte 0x5C repeated 64 times
		 * and text is the data being protected
		 */

		// start out by storing key in pads
		memset(k_ipad, 0, sizeof(k_ipad));
		memset(k_opad, 0, sizeof(k_opad));
		memcpy(k_ipad, ckey, keyLen);
		memcpy(k_opad, ckey, keyLen);

		// XOR key with ipad and opad values
		for (i = 0; i < kMaxKeyLen; i++)
		{
			k_ipad[i] ^= 0x36;
			k_opad[i] ^= 0x5C;
		}

		// perform inner MD5
		MD5Init(&context);							/* init context for 1st pass */
		MD5Update(&context, k_ipad, kMaxKeyLen);	/* start with inner pad */
		MD5Update(&context, cstring, stringLen);	/* then text of datagram */
		MD5Final(digest, &context);					/* finish up 1st pass */

		// perform outer MD5
		MD5Init(&context);							/* init context for 2nd pass */
		MD5Update(&context, k_opad, kMaxKeyLen);	/* start with outer pad */
		MD5Update(&context, digest, kDigestLen);	/* then results of 1st hash */
		MD5Final(digest, &context);					/* finish up 2nd pass */

		ConvertToHexString(digest, cstring);

		// return Unicode string
		result = MakeString((const char *) cstring);
	}

	return result;
}
