/*------------------------------------------------------------------------------
	Unicode text encoding functions.
------------------------------------------------------------------------------*/

#include "Objects.h"
#include "NewtonScript.h"
#include "Unicode.h"
#include "UnicodeMapper.h"


/*------------------------------------------------------------------------------
	C o n s t a n t s
------------------------------------------------------------------------------*/

enum
{
	kNullEncoding,
	kMailEncoding,
	kNumOfEncodings = 1
};

#define kNoIndex 0xFFFF


/*------------------------------------------------------------------------------
	T y p e s
------------------------------------------------------------------------------*/

typedef 	struct
{
	EncodingMap						map;
	ConvertFromUnicodeProcPtr	proc;
} FromUnicodeConverter;

typedef 	struct
{
	EncodingMap						map;
	ConvertToUnicodeProcPtr		proc;
} ToUnicodeConverter;

typedef struct
{
	FromUnicodeConverter fromUnicode;	// [kNumOfEncodings];
	ToUnicodeConverter toUnicode;			// [kNumOfEncodings];
} TTextEncoding;


/*------------------------------------------------------------------------------
	F u n c t i o n   P r o t o t y p e s
------------------------------------------------------------------------------*/

			Ref	TextEncodingData(RefArg rcvr);
static	Ref	GetTextEncodingDef(RefArg charset);
static	void	GetMappingInfo(void * data, EncodingMap * map, ProcPtr * mapFunction);
			long  AdjustLengthForEncoding(UniChar * inStr, long inLength, void * inData);
			long	ConvertFromUnicode(const UniChar * inStr, void * outStr, void * encoding, long numOfChars);
			long	ConvertToUnicode(const void * inStr, UniChar * outStr, void * encoding, long numOfChars);
static	long	subConvertFromUnicode(const UniChar * inStr, void * outStr, void * map, long numOfChars);
static	long	subConvertToUnicode(const void * inStr, UniChar * outStr, void * map, long numOfChars);
static	long	ConvertFromUnicodeFunc_Segmented16(const UniChar * inStr, void * outStr, void * map, long numOfChars);
static	long	ConvertToUnicodeFunc_Contiguous8(const void * inStr, UniChar * outStr, void * map, long numOfChars);
static	long	ConvertFromUnicodeFunc_MultiByte(const UniChar * inStr, void * outStr, void * map, long numOfChars);
static	long	ConvertToUnicodeFunc_MultiByte(const void * inStr, UniChar * outStr, void * map, long numOfChars);
static	long	ConvertFromUnicodeFunc_JIS(const UniChar * inStr, void * outStr, void * map, long numOfChars);
static	long	ConvertToUnicodeFunc_JIS(const void * inStr, UniChar * outStr, void * map, long numOfChars);
static	long	ConvertFromUnicodeFunc_UTF8(const UniChar * inStr, void * outStr, void * map, long numOfChars);
static	long	ConvertToUnicodeFunc_UTF8(const void * inStr, UniChar * outStr, void * map, long numOfChars);


/*------------------------------------------------------------------------------
	Install mapping-from-unicode tables into global data.

	Args:		data
					persistent data object
				charset
					symbol of charset

	Return:	nil
------------------------------------------------------------------------------*/

extern "C"
Ref InstallCharFromUnicodeEncoding(RefArg rcvr, RefArg charset)
{
	RefVar map;
	RefVar def(GetTextEncodingDef(charset));
	if (NOTNIL(def))
	{
		SetVariable(rcvr, SYM(mimeCharset), GetFrameSlot(def, SYM(symbol)));
		SetVariable(rcvr, SYM(mimeEncoding), GetFrameSlot(def, SYM(mimeEncoding)));

		map = GetFrameSlot(def, SYM(fromUnicode));
	}
	RefVar data(TextEncodingData(rcvr));
	WITH_LOCKED_BINARY(data, dataPtr)
		TTextEncoding * p = (TTextEncoding *) dataPtr;
		if (NOTNIL(map))
		{
			WITH_LOCKED_BINARY(map, mapPtr)
				GetMappingInfo(mapPtr, &p->fromUnicode.map, (ProcPtr *) &p->fromUnicode.proc);
			END_WITH_LOCKED_BINARY(map)
		}
		else
			p->fromUnicode.proc = subConvertFromUnicode;
	END_WITH_LOCKED_BINARY(data)

	return NILREF;
}


/*------------------------------------------------------------------------------
	Install mapping-to-unicode tables into global data.

	Args:		data
					persistent data object
				charset
					symbol of charset

	Return:	nil
------------------------------------------------------------------------------*/

extern "C"
Ref InstallCharToUnicodeEncoding(RefArg rcvr, RefArg charset)
{
	RefVar map;
	RefVar def(GetTextEncodingDef(charset));
	if (NOTNIL(def))
	{
		SetVariable(rcvr, SYM(mimeCharset), GetFrameSlot(def, SYM(symbol)));
		SetVariable(rcvr, SYM(mimeEncoding), GetFrameSlot(def, SYM(mimeEncoding)));

		map = GetFrameSlot(def, SYM(toUnicode));
	}
	RefVar data(TextEncodingData(rcvr));
	WITH_LOCKED_BINARY(data, dataPtr)
		TTextEncoding * p = (TTextEncoding *) dataPtr;
		if (NOTNIL(map))
		{
			WITH_LOCKED_BINARY(map, mapPtr)
				GetMappingInfo(mapPtr, &p->toUnicode.map, (ProcPtr *) &p->toUnicode.proc);
			END_WITH_LOCKED_BINARY(map)
		}
		else
			p->toUnicode.proc = subConvertToUnicode;
	END_WITH_LOCKED_BINARY(data)

	return NILREF;
}


/*------------------------------------------------------------------------------
	Get the applications global textEncodingData.

	Args:		rcvr
					the receiver, having a .textEncodingData object

	Return:	a binary object
------------------------------------------------------------------------------*/

Ref TextEncodingData(RefArg rcvr)
{
	RefVar dataTag(SYM(textEncodingData));
	RefVar data(GetVariable(rcvr, dataTag));
	if (ISNIL(data))
	{
		data = AllocateBinary(SYM(binary), sizeof(TTextEncoding));
		SetVariable(rcvr, dataTag, data);
	}
	return data;
}


/*------------------------------------------------------------------------------
	Get a text encoding def frame from the TextEncodings global variable.

	Args:		charset
					symbol of charset

	Return:	text encoding def
------------------------------------------------------------------------------*/

static Ref GetTextEncodingDef(RefArg charset)
{
	RefVar def;
	if (NOTNIL(charset))
		def = NSCallGlobalFn(SYM(GetGlobalVar), SYM(TextEncodings));
	if (NOTNIL(def))
		def = GetFrameSlot(def, charset);

	return def;
}


/*------------------------------------------------------------------------------
	Complete an EncodingMap record from a mapping table resource.

	Args:		data
					mapping table data
				map
					encoding map record to be completed
				mapFunction
					address of ProcPtr to be set

	Return:	void
------------------------------------------------------------------------------*/

static void GetMappingInfo(void * data, EncodingMap * map, ProcPtr * mapFunction)
{
	UShort * table = (UShort *) data;

	map->mapFormatID = *table++;
	map->tableSize = *table++;
	map->revision = *table++;
	map->tableInfo = *table++;

	switch (map->mapFormatID)
	{
	case 0:  // 8-bit -> Unicode
		map->charCodes = table;
		*mapFunction = (ProcPtr) ConvertToUnicodeFunc_Contiguous8;
		break;

	case 1:  // multi-byte -> Unicode
		map->startCodes = (UniChar *) table;
		table += 128;  // 256 / sizeof(UShort)
		map->endCodes = (UniChar *) table;
		table += 128;  // 256 / sizeof(UShort)
		map->codeDeltas = (SShort *) table;
		table++;
		map->charCodes = table;
		*mapFunction = (ProcPtr) ConvertToUnicodeFunc_MultiByte;
		break;

	case 2:  // JIS (ISO 2022-JP) -> Unicode
		map->startCodes = (UniChar *) table;
		map->endCodes = (UniChar *) table;
		table += 64;  // 128 / sizeof(UShort)
		map->codeDeltas = (SShort *) table;
		table++;
		map->charCodes = table;
		*mapFunction = (ProcPtr) ConvertToUnicodeFunc_JIS;
		break;

	case 3:  // UTF-8 -> Unicode
		*mapFunction = (ProcPtr) ConvertToUnicodeFunc_UTF8;
		break;

	case 4:  // Unicode -> 8-bit
		map->endCodes = (UniChar *) table;
		table += map->tableSize;
		map->startCodes = (UniChar *) table;
		table += map->tableSize;
		map->codeDeltas = (SShort *) table;
		table += map->tableSize;
		map->charCodes = table;
		*mapFunction = (ProcPtr) ConvertFromUnicodeFunc_Segmented16;
		break;

	case 5:  // Unicode -> multi-byte
	case 6:  // Unicode -> JIS (ISO 2022-JP)
		{
			UShort indexTableSize = *table++;
			map->startCodes = (UniChar *) table;			// actually index table
			table += indexTableSize;
			map->charCodes = table;
			table += map->tableSize;
			map->endCodes = (UniChar *) table;				// actually gap table
		}
		if (map->mapFormatID == 5)
			*mapFunction = (ProcPtr) ConvertFromUnicodeFunc_MultiByte;
		else
			*mapFunction = (ProcPtr) ConvertFromUnicodeFunc_JIS;
		break;

	case 7:  // Unicode -> UTF-8
		*mapFunction = (ProcPtr) ConvertFromUnicodeFunc_UTF8;
		break;

	default:
		*mapFunction = (ProcPtr) NULL;
	}
}


/*------------------------------------------------------------------------------
	Adjust the length of an 8-bit binary string to cope with multi-byte
	encoding.

	Args:		inStr
					the Unicode source string
				ioLength
					the number of UniChars in the Unicode source string
				inData
					TTextEncoding context

	Return:	long
					number of chars to allocate
------------------------------------------------------------------------------*/

long AdjustLengthForEncoding(UniChar * inStr, long ioLength, void * inData)
{
	long  expansion = ((TTextEncoding *) inData)->fromUnicode.map.tableInfo >> 8;
	if (expansion == 1)
		ioLength *= sizeof(UniChar);
	else if (expansion == 2)
		ioLength = 64 + ioLength * sizeof(UniChar) + (ioLength >> 5);
	else if (expansion == 3)
		for (long i = ioLength; i >= 0; i--, inStr++)
			if (*inStr > 0x0800)
				ioLength += 2;
			else if (*inStr > 0x0080)
				ioLength++;
	return ioLength;
}


/*------------------------------------------------------------------------------
	Convert from Unicode to 8-bit character string.
	NUL terminate the string.

	Args:		inStr
					a Unicode string
				outStr
					an 8-bit character string
				encoding
					char encoding data
				numOfChars
					number of characters to encode

	Return:	long
					number of characters actually converted
------------------------------------------------------------------------------*/

long xConvertFromUnicode(const UniChar * inStr, void * outStr, void * encoding, long numOfChars)
{
	if (encoding == NULL || ((TTextEncoding *) encoding)->fromUnicode.proc == NULL)
		return subConvertFromUnicode(inStr, outStr, NULL, numOfChars);
	return ((TTextEncoding *) encoding)->fromUnicode.proc(inStr, outStr, &((TTextEncoding *) encoding)->fromUnicode.map, numOfChars);
}


static long subConvertFromUnicode(const UniChar * inStr, void * outStr, void * map, long numOfChars)
{
	UniChar ch;
	unsigned char * p = (unsigned char *) outStr;
	for ( ; numOfChars > 0 && (ch = *inStr++) != kEndOfUnicodeString; numOfChars--)
		*p++ = (ch >= 0x80) ? kNoTranslationChar : ch;
	*p = kEndOfCharString;
	return p - (unsigned char *) outStr;
}


/*------------------------------------------------------------------------------
	Convert to Unicode from 8-bit character string.
	NUL terminate the string.

	Args:		inStr
					an 8-bit character string
				outStr
					a Unicode string
				encoding
					char encoding data
				numOfChars
					number of characters to encode

	Return:	long
					number of characters actually converted
------------------------------------------------------------------------------*/

long xConvertToUnicode(const void * inStr, UniChar * outStr, void * encoding, long numOfChars)
{
	if (encoding == NULL || ((TTextEncoding *) encoding)->toUnicode.proc == NULL)
		return subConvertToUnicode(inStr, outStr, NULL, numOfChars);
	return ((TTextEncoding *) encoding)->toUnicode.proc(inStr, outStr, &((TTextEncoding *) encoding)->toUnicode.map, numOfChars);
}


static long subConvertToUnicode(const void * inStr, UniChar * outStr, void * map, long numOfChars)
{
	unsigned char ch;
	const unsigned char * p = (const unsigned char *) inStr;
	UniChar * s = outStr;
	for ( ; numOfChars > 0 && (ch = *p++) != kEndOfCharString; numOfChars--)
		if (ch != 0x0A)
			*s++ = ch;
	*s = kEndOfUnicodeString;
	return s - outStr;
}


/*------------------------------------------------------------------------------
	Convert from Unicode to 8-bit character string.
	NUL terminate the string.

	Args:		inStr
					a Unicode string
				outStr
					an 8-bit character string
				map
					encoding map
				numOfChars
					number of characters to encode
					Encoding is terminated early if NUL is encountered in inStr.

	Return:	long
					number of characters actually converted
------------------------------------------------------------------------------*/

static long ConvertFromUnicodeFunc_Segmented16(const UniChar * inStr, void * outStr, void * map, long numOfChars)
{
	UniChar unic;
	unsigned char ch;
	unsigned char * p = (unsigned char *) outStr;
	for ( ; numOfChars > 0 && (unic = *inStr++) != kEndOfUnicodeString; numOfChars--)
	{
		ch = kNoTranslationChar;
		for (int i = 0; i < ((EncodingMap *) map)->tableSize; i++)
		{
			if ((unic >= ((EncodingMap *) map)->startCodes[i])
			 && (unic <= ((EncodingMap *) map)->endCodes[i]))
			{
				ch = ((unsigned char *) ((EncodingMap *) map)->charCodes)[unic + ((EncodingMap *) map)->codeDeltas[i]];
				break;
			}
		}
		*p++ = ch;
	}
	*p = kEndOfCharString;
	return p - (unsigned char *) outStr;
}


/*------------------------------------------------------------------------------
	Convert to Unicode from 8-bit character string.
	NUL terminate the string.
	Also strip LFs - Newton does not use em.

	Args:		inStr
					an 8-bit character string
				outStr
					a Unicode string
				map
					encoding map
				numOfChars
					number of characters to encode
					Encoding is terminated early if NUL is encountered in inStr.

	Return:	long
					number of characters actually converted
------------------------------------------------------------------------------*/

static long ConvertToUnicodeFunc_Contiguous8(const void * inStr, UniChar * outStr, void * map, long numOfChars)
{
	unsigned char ch;
	const unsigned char * p = (const unsigned char *) inStr;
	UniChar * s = outStr;
	for ( ; numOfChars > 0 && (ch = *p++) != kEndOfCharString; numOfChars--)
		if (ch != 0x0A)
			*s++ = ((UniChar *) ((EncodingMap *) map)->charCodes)[ch];
	*s = kEndOfUnicodeString;
	return s - outStr;
}

#pragma mark -

/*------------------------------------------------------------------------------
	Convert from Unicode to UTF-8 character string.
	NUL terminate the string.

	Args:		inStr
					a Unicode string
				outStr
					a UTF-8 multi-byte character string
				map
					encoding map - ignored, UTF-8 is mapped directly to Unicode
				numOfChars
					number of characters to encode
					Encoding is terminated early if NUL is encountered in inStr.

	Return:	long
					number of characters actually converted
------------------------------------------------------------------------------*/

static long	ConvertFromUnicodeFunc_UTF8(const UniChar * inStr, void * outStr, void * map, long numOfChars)
{
	UniChar unic;
	unsigned char ch;
	unsigned char * p = (unsigned char *) outStr;
	for ( ; numOfChars > 0 && (unic = *inStr++) != kEndOfUnicodeString; numOfChars--)
	{
		if (unic < 0x0080)
			*p++ = unic;
		else if (unic < 0x0800)
		{
			*p++ = 0xC0 | (unic >> 6);
			*p++ = 0x80 | (unic & 0x3F);
		}
		else
		{
			*p++ = 0xE0 | (unic >> 12);
			*p++ = 0x80 | ((unic >> 6) & 0x3F);
			*p++ = 0x80 | (unic & 0x3F);
		}
	}
	*p = kEndOfCharString;
	return p - (unsigned char *) outStr;
}

/*------------------------------------------------------------------------------
	Convert to Unicode from UTF-8 multi-byte character string.
	NUL terminate the string.
	Also strip LFs - Newton does not use em.

	Args:		inStr
					a multi-byte character string
				outStr
					a Unicode string
				map
					encoding map - ignored, Unicode is mapped directly to UTF-8
				numOfChars
					length of inStr
					Encoding is terminated early if NUL is encountered in inStr.

	Return:	long
					number of Unicode characters actually converted
------------------------------------------------------------------------------*/

static long	ConvertToUnicodeFunc_UTF8(const void * inStr, UniChar * outStr, void * map, long numOfChars)
{
	unsigned char ch;
	unsigned char chBits;
	const unsigned char * p = (const unsigned char *) inStr;
	UniChar utf8ch;
	UniChar * s = outStr;
	for ( ; numOfChars > 0 && (ch = *p++) != kEndOfCharString; numOfChars--)
	{
		chBits = ch & 0xE0;
		if (chBits == 0xE0)
		{
			utf8ch = (ch & 0x0F) << 12;
			if (--numOfChars > 0 && (ch = *p) != kEndOfCharString)
			{
				p++;
				utf8ch |= ((ch & 0x3F) << 6);
				if (--numOfChars > 0 && (ch = *p) != kEndOfCharString)
				{
					p++;
					*s++ = utf8ch | (ch & 0x3F);
				}
			}
		}
		else if (chBits == 0xC0)
		{
			utf8ch = (ch & 0x1F) << 6;
			if (--numOfChars > 0 && (ch = *p) != kEndOfCharString)
			{
				p++;
				*s++ = utf8ch | (ch & 0x3F);
			}
		}
		else if (ch != 0x0A)
			*s++ = ch;
		}
	*s = kEndOfUnicodeString;
	return s - outStr;
}

#pragma mark -

/*------------------------------------------------------------------------------
	Convert from Unicode to multi-byte character string.
	NUL terminate the string.

	Args:		inStr
					a Unicode string
				outStr
					an 8-bit multi-byte character string
				map
					encoding map
				numOfChars
					number of characters to encode
					Encoding is terminated early if NUL is encountered in inStr.

	Return:	long
					number of characters actually converted
------------------------------------------------------------------------------*/

static long ConvertFromUnicodeFunc_MultiByte(const UniChar * inStr, void * outStr, void * map, long numOfChars)
{
	long					indexShift = ((EncodingMap *) map)->tableInfo & 0x0F;
	long					indexMask = (1 << indexShift) - 1;
	long					codeOffset, blockOffset, delta;
	unsigned short *  codePtr;
	unsigned char *	gapPtr;
	UniChar				unic;
	unsigned short		mbch;
	unsigned char		ch;
	unsigned char *	p = (unsigned char *) outStr;
	for ( ; numOfChars > 0 && (unic = *inStr++) != kEndOfUnicodeString; numOfChars--)
	{
		if (unic < 0x80)
			ch = (unsigned char) unic;
		else
		{
			mbch = 0xFFFD;
			codeOffset = ((unsigned short *) ((EncodingMap *) map)->startCodes)[unic >> indexShift];
			if (codeOffset != kNoIndex)
			{
				codePtr = (unsigned short *) ((EncodingMap *) map)->charCodes + codeOffset;
				gapPtr = (unsigned char *) ((EncodingMap *) map)->endCodes + codeOffset;

				// loop until we find the right index
				for (blockOffset = unic & indexMask; blockOffset > 0; blockOffset -= delta)
				{
					codePtr++;
					delta = *gapPtr++;
				}
				if (blockOffset == 0)
					mbch = *codePtr;
			}
			*p++ = mbch >> 8;
			ch = mbch & 0xFF;
		}
		*p++ = ch;
	}
	*p = kEndOfCharString;
	return p - (unsigned char *) outStr;
}


/*------------------------------------------------------------------------------
	Convert to Unicode from multi-byte character string.
	NUL terminate the string.
	Also strip LFs - Newton does not use em.
	In this context the EncodingMap startCodes and endCodes are
	redefined as tables of valid byte 1 & 2 values of a multi-byte code.

	Args:		inStr
					a multi-byte character string
				outStr
					a Unicode string
				map
					encoding map
				numOfChars
					length of inStr
					Encoding is terminated early if NUL is encountered in inStr.

	Return:	long
					number of Unicode characters actually converted
------------------------------------------------------------------------------*/

static long ConvertToUnicodeFunc_MultiByte(const void * inStr, UniChar * outStr, void * map, long numOfChars)
{
	unsigned char  ch1, ch2;
	unsigned long  off1, off2;
	unsigned short mbch;
	unsigned long  numOfCodes2 = *((EncodingMap *) map)->codeDeltas;
	const unsigned char * p = (const unsigned char *) inStr;
	UniChar * s = outStr;
	for ( ; numOfChars > 0 && (ch1 = *p++) != kEndOfCharString; numOfChars--)
	{
		if (ch1 != 0x0A)
		{
			if ((off1 = ((UChar *) ((EncodingMap *) map)->startCodes)[ch1]) == 0)
				// its a 'normal' character
				*s++ = ch1;
			else if (off1 == 1)
				 // its a Shift JIS (M$) hack for half-width katakana
				*s++ = 0xFF61 - 0xA1 + ch1;
			else if (((ch2 = *p) != kEndOfCharString)
					&& p++, numOfChars--, (off2 = ((UChar *) ((EncodingMap *) map)->endCodes)[ch2]) != 0
					&& (mbch = (ch1 - off1) * numOfCodes2 + (ch2 - off2)) < ((EncodingMap *) map)->tableSize)
				// its a valid DBCS character
				*s++ = ((UniChar *) ((EncodingMap *) map)->charCodes)[mbch];
			else
				// its rubbish
				*s++ = kNoTranslationChar;
		}
	}
	*s = kEndOfUnicodeString;
	return s - outStr;
}

#pragma mark -

/*------------------------------------------------------------------------------
	Convert from Unicode to JIS (ISO 2022-JP) character string.
	NUL terminate the string.

	Args:		inStr
					a Unicode string
				outStr
					a 7-bit ISO 2022-JP character string
				map
					encoding map
				numOfChars
					number of characters to encode
					Encoding is terminated early if NUL is encountered in inStr.

	Return:	long
					number of characters actually converted
------------------------------------------------------------------------------*/

enum
{
	kJISASCII,
	kJIS0201,
	kJIS0208,

	kJISEscape,
	kJISSingle,
	kJISDouble
};

// Escape string matrix, from/to mode
const char * const kEscapeStr[3][3] =
{
				 /*ASCII*/		/*Roman*/	/*JIS0208*/
/*ASCII*/	{  NULL,			"\033(J",	"\033$B" },
/*Roman*/	{  NULL,			NULL,			"\033$B" },
/*JIS0208*/ {  "\033(B",	"\033(J",	NULL		}
};

#define kEscapeStrLen 3


static long ClassifyJIS(UniChar unic)
{
	long  classification = kJISASCII;
	if (unic >= 0x80)
	{
		if (unic == 0x00A5			// Yen
		 || unic == 0x203E			// overline
		 || (unic >= 0xFF61 && unic <= 0xFF9E))	// half-width katakana
			classification = kJIS0201;
		else
			classification = kJIS0208;
	}
	return classification;
}


static long ConvertFromUnicodeFunc_JIS(const UniChar * inStr, void * outStr, void * map, long numOfChars)
{
	long					indexShift = ((EncodingMap *) map)->tableInfo & 0x0F;
	long					indexMask = (1 << indexShift) - 1;
	long					codeOffset, blockOffset, delta;
	unsigned short *  codePtr;
	unsigned char *	gapPtr;
	const char *		transition;
	long					newState, state = kJISASCII;
	UniChar				unic;
	unsigned short		mbch;
	unsigned char		ch;
	unsigned char *	p = (unsigned char *) outStr;
	for ( ; numOfChars > 0 && (unic = *inStr++) != kEndOfUnicodeString; numOfChars--)
	{
		newState = ClassifyJIS(unic);
		if ((transition = kEscapeStr[state][newState]) != NULL)
		{
			memcpy(p, transition, kEscapeStrLen);
			p += kEscapeStrLen;
			state = newState;
		}

		switch (state)
		{
		case kJISASCII:
			ch = unic;
			break;

		case kJIS0201:
			if (unic == 0x00A5)
				ch = '\\';			// Yen -> backslash
			else if (unic == 0x203E)
				ch = '~';			// overline -> tilde
			else if (unic >= 0xFF61 && unic <= 0xFFEF)
				ch = unic - 0xFF61 + 0xA1;	// half-width katakana
			else
				ch = unic;
			break;

		case kJIS0208:
			mbch = 0xFFFD;
			codeOffset = ((unsigned short *) ((EncodingMap *) map)->startCodes)[unic >> indexShift];
			if (codeOffset != kNoIndex)
			{
				codePtr = (unsigned short *) ((EncodingMap *) map)->charCodes + codeOffset;
				gapPtr = (unsigned char *) ((EncodingMap *) map)->endCodes + codeOffset;

				// loop until we find the right index
				for (blockOffset = unic & indexMask; blockOffset > 0; blockOffset -= delta)
				{
					codePtr++;
					delta = *gapPtr++;
				}
				if (blockOffset == 0)
					mbch = *codePtr;
			}
			*p++ = mbch >> 8;
			ch = mbch & 0xFF;
			break;
		}
		*p++ = ch;
	}
	// leave string in single-byte state
	if (state == kJIS0208)
	{
		memcpy(p, kEscapeStr[kJIS0208][kJISASCII], kEscapeStrLen);
		p += kEscapeStrLen;
	}
	*p = kEndOfCharString;
	return p - (unsigned char *) outStr;
}


/*------------------------------------------------------------------------------
	Convert to Unicode from JIS (ISO 2022-JP) encoded character string.
	NUL terminate the string.
	Also strip LFs - Newton does not use em.
	In this context the EncodingMap startCodes and endCodes are
	redefined as tables of valid byte 1 & 2 values of a multi-byte code.

	Args:		inStr
					a multi-byte character string
				outStr
					a Unicode string
				map
					encoding map
				numOfChars
					length of inStr
					Encoding is terminated early if NUL is encountered in inStr.

	Return:	long
					number of Unicode characters actually converted
------------------------------------------------------------------------------*/

static long ConvertToUnicodeFunc_JIS(const void * inStr, UniChar * outStr, void * map, long numOfChars)
{
	unsigned char  ch1, ch2;
	unsigned long  off1, off2;
	unsigned short mbch;
	unsigned long  numOfCodes2 = *((EncodingMap *) map)->codeDeltas;
	long		state = kJISASCII;
	UniChar  unic;
	const unsigned char * p = (const unsigned char *) inStr;
	UniChar * s = outStr;
	for ( ; numOfChars > 0 && (ch1 = *p++) != kEndOfCharString; numOfChars--)
	{
		if (ch1 != 0x0A)
		{
			if (ch1 == 0x1B)
				state = kJISEscape;
			else switch (state)
			{
			case kJISASCII:
				*s++ = ch1;
				break;

			case kJISEscape:
				if (ch1 == '(')
					state = kJISSingle;	// switching to single-byte character set
				else if (ch1== '$')
					state = kJISDouble;	// switching to double-byte character set
				else
					state = kJISASCII;	// unknown escape
				break;

			case kJISSingle:
				if (ch1 == 'B')
					state = kJISASCII;	// switching to ASCII
				else if (ch1== 'J')
					state = kJIS0201;		// switching to Roman
				else
					state = kJISASCII;	// unknown escape
				break;

			case kJISDouble:
				if (ch1 == '@'
				 || ch1 == 'B')
					state = kJIS0208;		// switching to JIS X 0208
				else
					state = kJISASCII;	// unknown escape
				break;

			case kJIS0201:
				if (ch1 == '\\')
					unic = 0x00A5;			// backslash -> Yen
				else if (ch1 == '~')
					unic = 0x203E;			// tilde -> overline
				// JIS8 codes, as in libiconv
				else if (ch1 >= 0xA1 && ch1 <= 0xDF)
					unic = 0xFF61 - 0xA1 + ch1;	// half-width katakana
				else
					unic = ch1;
				*s++ = unic;
				break;

			case kJIS0208:
				if ((off1 = ((UChar *) ((EncodingMap *) map)->startCodes)[ch1]) != 0
					&& (ch2 = *p) != kEndOfCharString
					&& p++, numOfChars--, (off2 = ((UChar *) ((EncodingMap *) map)->endCodes)[ch2]) != 0
					&& (mbch = (ch1 - off1) * numOfCodes2 + (ch2 - off2)) < ((EncodingMap *) map)->tableSize)
					// its a valid DBCS character
					*s++ = ((UniChar *) ((EncodingMap *) map)->charCodes)[mbch];
				else
					// its rubbish, but let ch1 thru in case encoder didnt escape it correctly
					*s++ = ch1;		// kNoTranslationChar;
				break;
			}
		}
	}
	*s = kEndOfUnicodeString;
	return s - outStr;
}
